(* marioxcc_tactics.sml - Tactics and utilities for HOL4

Copyright © Mario Castelán Castro.

This work is licensed as free software under the GNU General Public License
either version 3 or any later version published by the Free Software
Foundation. See <https://www.gnu.org/licenses/gpl.html>.

As an additional permission (see section 7 of the GNU General Public
License, version 3) you may convey copies of this work or a work based on
this work without a copy of the license. The other requirements of the
license remain in effect. *)

structure marioxcc_tactics :> marioxcc_tactics =
struct

open HolKernel Parse boolLib bossLib List;

fun wrap x = [x];

fun parse_in_goal_context (g as (asm, _:term)) term =
  Parse.parse_in_context (Term.free_varsl asm) term;

(* Split a goal of the form (“P ∧ Q”, “P ∧ Q ∧ R”, ...) into one subgoal
for each conjunct. The subgoals are in the same order as in the goal. This
tactic does not recursively splits conjuncts.

Example: nconj_tac Γ ?- a ∧ (b ∧ c) ∧ d yields

Γ ?- a
Γ ?- (b ∧ c)
Γ ?- d. *)
val nconj_tac:tactic =
  fn (g as (asm, concl)) =>
    let
      val conj_list = spine_binop (total dest_conj) concl
      val err = mk_HOL_ERR "topologyScript" "NSTRIP_TAC" "Not a conjuction"
      val _ = if length conj_list <= 1 then raise err else ()
      val subgoals = map (fn elem => (asm, elem)) conj_list
      fun validation (thm, NONE) = SOME thm
        | validation (thm, SOME prev_thm) = SOME (CONJ thm prev_thm)
    in
      (subgoals, valOf o (foldr validation NONE))
    end

(* A future version of “strip_all” will be able to strip restricted
quantifiers and will not cause case split on disjunctions. Proofs should
take this into account. *)
val strip_all = rpt strip_tac;

(* Apply the given tactic and then performs an α-conversion of the
conclusion of the goal to the user-provided term (passed as a quotation and
parsed in the context of the original goal).

Fails if the tactic does not produce a single subgoal or if the conclussion
of the subgoal is not α-equivalent to the user-provided term.

The name “trfm” is an abbreviation of “transform”. A call “trfm quot tac”
can be interpreted as a declaration that “tac” transforms the conclussion
of the subgoal into “quot”. *)
val trfm:(term quotation -> tactic -> tactic) =
  fn quot => fn tac => fn g as (asm, concl) =>
    let
      val err = mk_HOL_ERR "topologyScript" "trfm"
      val term = parse_in_goal_context g quot
      val (subgoals, subvalidation) = tac g
      val _ =
        if length subgoals <> 1 then
        raise err "Not just one subgoal produced"
        else ()
      val (_, new_concl) = hd subgoals
      val _ =
        if not (aconv term new_concl)
        then raise err "Not alpha equivalent"
        else ()
    in
      ([(asm, term)], subvalidation)
    end

(* Continues the proof with the quotation as the new goal, parsed in the
context of the assumptions. The quotation must be α-equivalent to
the goal; otherwise it fails. The provided tactic must solve the goal.

This tactical can be used to declare what the goal is in declarative-style
proofs. *)
fun must_prove quot tac = trfm quot ALL_TAC >- tac;

(* Calling “lemma t tac Γ ?- c” returns the subgoals generated by the
application of “tac” (if any) followed by the goal “Γ,t ?- c”. “t” is
parsed to a term using the free variables of Γ as context.

When “tac” solves “Γ ?- t” outright, then “lemma t tac Γ ?- c” returns
Γ,t ?- c as the only subgoal.

This tactical is like “by”, but it does not perform any processing of the
result once it is proved; it is added to assumptions as-is. *)
val lemma:(term frag list -> tactic -> tactic) =
  fn quot => fn tac => fn g as (asm, concl) =>
    let
      val term:term = Parse.parse_in_context (Term.free_varsl asm) quot
      val (subgoals:goal list, subvalidation) = tac (asm, term)
      val validation =
        fn thm_list =>
          (* The first “length subgoals” theorems achieve the subgoals
          generated by the user-provided tactic; the last theorem achieves
          the enhancement of the original goal with the user-provided
          lemma. *)
          PROVE_HYP
            (subvalidation (take (thm_list, length subgoals)))
            (last thm_list)
    in
      (subgoals @ [(term :: asm, concl)], validation)
    end

(* Use the assumption at the top of the stack achieves the goal.

This tactic is intended to be used when the goal has been solved and added
to the assumption stack with the tactical “lemma”. *)
val done:tactic = POP_ASSUM ACCEPT_TAC;

(* Precursor for “now” and “try_now”. *)
val gen_now:(bool -> term frag list -> tactic -> tactic) =
  fn strict => fn quot => fn tac => fn g as (asm, concl) =>
    if null asm then
      raise mk_HOL_ERR "topologyScript" "now" "No assumptions"
    else
      let
        val term:term = Parse.parse_in_context (Term.free_varsl asm) quot
        val (subgoals, subvalidation) = tac (asm, term)
        val _ =
          if strict andalso not (null subgoals)
          then raise mk_HOL_ERR "topologyScript" "now" "Unsolved subgoals"
          else ()
        val new_goal = (term :: tl asm, concl)
        fun validation thm_list =
          PROVE_HYP
            (subvalidation (take (thm_list, length subgoals)))
            (last thm_list)
      in
        (subgoals @ [new_goal], validation)
      end

(* Suppose that we have a tactic “tac” that proves Γ,a ?- t outright, then
“now tac t” turns the goal Γ,a ?- p into Γ,t ?- p.

This tactical is intended to supplement “by” in proofs in partially or
wholly declarative style. Consider a chain of tactics of the form

(1) P1 by Tac1
(2) >> P2 by Tac2
(n) >> PN by TacN

where each PM is only used to prove P(M+1). The “by” tactical keeps all the
intermediate results as assumptions even after they are no longer
needed. The “now” tactical instead replaces each intermediate result by the
following one. *)
val now = gen_now true;

(* Like “now”, but if the user-provided tactic does not solve the goal then
the subgoals are returned in the same order and before the goal that “new”
would have returned. *)
val now_ns = gen_now false;

(* Pop the all assumptions that are α-equivalent to the passed quotation
(parsed in context) and turn the goal into an implication with the passed
quotation as the antecedent. Fails if no suitable assumption is found. *)
val use:(term quotation -> tactic) =
  fn quot => fn g as (asm, concl) =>
    let
      val term = parse_in_goal_context g quot
      val rest_asm = op_set_diff aconv asm [term]
      val _ =
        if null rest_asm
        then raise mk_HOL_ERR "topologyScript" "use" "No match"
        else ()
    in
      MP_TAC (ASSUME term) (rest_asm, concl)
    end

val use_prev = POP_ASSUM MP_TAC;

(* Continue the proof with the user-provided terms (passed as a quotation
and parsed in the context of the original goal) as the only assumptions, in
the given order. Each user-provided term must be α-equivalent to at least
one assumption in the original goal. *)
val with_assuml:(term quotation list -> tactic) =
  fn quot => fn g as (asm, concl) =>
    let
      val terms = map (parse_in_goal_context g) quot
      val err = mk_HOL_ERR "topologyScript" "use" "No match"
      fun validate_assum t =
        if exists (aconv t) asm
        then ()
        else raise err
    in
      app validate_assum terms;
      ([(terms, concl)], hd)
    end

val with_assum = with_assuml o wrap;
fun suffices quot tac = quot suffices_by tac;
val bsimp = SIMP_TAC bool_ss;
val asm_bsimp = ASM_SIMP_TAC bool_ss;
val full_bsimp = FULL_SIMP_TAC bool_ss;

end (* struct *)
